/// CaveGame - A procedural 2D platformer sandbox.
/// Created by Josh O'Leary @ Redacted Software, 2020-2025
/// Contact: josh@redacted.cc
/// Contributors: william@redacted.cc maxi@redacted.cc
/// This work is dedicated to the public domain.

/// @file Camera2D.hpp
/// @desc A 2-D camera controller class.
/// @edit 12/2/2024
/// @auth Josh O'Leary
#pragma once

#include <J3ML/Geometry/AABB2D.hpp>

namespace CaveGame::Client
{
    /// 2D Camera with sophisticated controls.
    class Camera2D
    {
    public:
        /// The default constructor does not initialize any values.
        Camera2D();

        // TODO: Implement camera shaking for explosions and other effects.
        void ApplyShake(const Vector2& impulse);
        /// Performs one iteration of the camera's internal logic.
        /// @param elapsed The delta-time (in seconds) between this iteration and the last.
        void Update(float elapsed);
        /// Returns the bounds of the screen-viewport, in gameworld-coordinates, without scaling by the camera's zoom.
        AABB2D Viewport() const;
        /// Returns the bounds of the screen-viewport, in gameworld-coordinates, with scaling by the camera's zoom.
        AABB2D ScaledViewport() const;

        AABB2D ScaledViewportOversized(float oversize) const;
        /// Returns the target coordinates of this camera. The camera will move over time to reach this point.
        Vector2 Target() const;
        /// Sets the target coordinates of this camera. The camera will move over time to reach this point.
        void Target(const Vector2& pos);
        /// Returns the constant used to linearly interpolate the camera's position towards it's goal.
        float PositionSmoothingFactor() const;
        /// Sets the constant used to linearly interpolate the camera's position towards it's goal.
        void PositionSmoothingFactor(float smoothing);
        /// Returns whether the camera should smoothly track it's target, or be locked on.
        bool PositionSmoothingEnabled() const;
        /// Sets whether the camera should smoothly track it's target.
        void PositionSmoothingEnabled(bool enabled);
        /// Returns the amount of zoom in the camera transformation.
        [[nodiscard]] float Zoom() const;
        /// Sets the amount of zoom in the camera transformation.
        void Zoom(float val);

        /// Enables free camera movement controls.
        void SetFreeCam(bool freeCam);

        /// Returns whether free camera controls are enabled.
        bool GetFreeCam() const;

        /// Zooms the camera in by the given amount.
        void ZoomIn(float amt);

        /// Zooms the camera out by the given amount.
        void ZoomOut(float amt);

        /// Returns the amount of rotation (in degrees) in the camera transformation.
        [[nodiscard]] float Rotation() const;

        /// Sets the amount of rotation (in degrees) in the camera transformation.
        void Rotation(float val);

        /// Rotates the camera by the given amount (in degrees).
        void Rotate(float angles);

        /// Converts top-left origin screen coordinates into world-space coordinates, with respect to zoom scale.
        Vector2 ScreenToWorld(const Vector2& device_coordinates) const;

        /// Converts world-space coordinates into top-left origin screen coordinates, with respect to zoom scale.
        Vector2 WorldToScreen(const Vector2& position) const;

        /// Returns the current cartesian coordinates of the camera transformation.
        Vector2 Position() const;

        /// Sets the current cartesian coordinates of the camera transformation.
        void Position(const Vector2& newPos);

        /// Returns the transformation matrix constructed from the camera's position, scale (zoom), and rotation.
        Matrix4x4 Transform() const;


        /// Moves the camera to the left at a given rate times the base camera speed.
        void MoveLeft(float rate = 1.f);
        /// Moves the camera to the right at a given rate times the base camera speed.
        void MoveRight(float rate = 1.f);
        /// Moves the camera up at a given rate times the base camera speed.
        void MoveUp(float rate = 1.f);
        /// Moves the camera down at a given rate times the base camera speed.
        void MoveDown(float rate = 1.f);

        /// Update the camera class with the program's current window resolution.
        void PassViewportSize(const Vector2& size);


        [[nodiscard]] Vector2 TopLeft() const;

        /// Returns the gameworld-coordinates at the top-left of the screen, with scaling taken into account.
        [[nodiscard]] Vector2 ScaledTopLeft() const;

        /// Returns the world-space coordinates at the focal point of the camera. Generally, this should also be the
        /// world-space coordinates of the center of the screen.
        [[nodiscard]] Vector2 Center() const;

        /// Returns the "radius" of the viewport. (size / 2.f)
        [[nodiscard]] Vector2 HalfSizeOffset() const;

        /// Returns the "radius" of the viewport, scaled by the camera zoom.
        [[nodiscard]] Vector2 ScaledHalfSizeOffset() const;

        /// Returns the base movement speed of the camera.
        float MoveSpeed() const;

        /// Sets the base movement speed of the camera.
        void MoveSpeed(float value);
    protected:
        /// Process user-input for freecam.
        void UpdateFreecamControls(float delta);
    protected:
        Vector2 velocity;
        Vector2 position;
        Vector2 final_pos;
        float zoom = 1.f;
        float rotation = 0.f;
        Vector2 target = {0, 0};
        Vector2 viewport_size = {800, 600};
        float speed = 0.5f;
        float lerp_factor = 10.f;
        bool lerp_to_target = true;
        bool freecam = true;
    private:

    };
}