#include <format>
#include <Client/LocalWorld.hpp>
#include <JGL/JGL.h>
#include <Core/Data.hpp>
#include <fstream>
#include <Core/Loggers.hpp>
#include <JUI/Widgets/Scene.hpp>

namespace CaveGame::Client {

    LocalWorld::LocalWorld(const std::string& world_name, int seed, bool overwrite)
        : World(world_name, seed, overwrite) {
        Logs::Info("Spinning up ChunkServerThread");
        chunk_thread = std::thread(&LocalWorld::ChunkServerThread, this);
        chunk_thread.detach();
        //hud = nullptr;
    }

    void LocalWorld::PostInit() {
        //font = JGL::Font("assets/fonts/Jupiteroid.ttf");
    }

    void LocalWorld::DrawChunkGrid() const {
        Vector2 viewport_topleft = camera.ScaledViewport().minPoint;
        Vector2 viewport_bottomright = camera.ScaledViewport().maxPoint;

        int nearest_grid_left = Math::Floor(viewport_topleft.x / Core::Chunk::ChunkSize);
        int nearest_grid_right = Math::Floor(viewport_bottomright.x / Core::Chunk::ChunkSize);

        for (int x = nearest_grid_left; x <= nearest_grid_right; x++) {
            auto top = Vector2(x * Core::Chunk::ChunkSize, viewport_topleft.y);
            auto bottom = Vector2(x * Core::Chunk::ChunkSize, viewport_bottomright.y);
            JGL::J2D::DrawLine(Colors::Red, top, bottom);
        }

        int nearest_grid_top = Math::Floor(viewport_topleft.y / Core::Chunk::ChunkSize);
        int nearest_grid_bottom = Math::Floor(viewport_bottomright.y / Core::Chunk::ChunkSize);

        for (int y = nearest_grid_top; y <= nearest_grid_bottom; y++) {
            auto left = Vector2(viewport_topleft.x, y * Core::Chunk::ChunkSize);
            auto right = Vector2(viewport_bottomright.x, y * Core::Chunk::ChunkSize);
            JGL::J2D::DrawLine(Colors::Red, left, right);
        }
    }


    void LocalWorld::CheckCachedChunkSprites()
    {
        for (auto& [chunk_pos, chunk] : loaded_chunks) {
            if (IsChunkCellWithinViewport(chunk_pos))
            {
                // No rendertarget for this chunk.
                if (!cached_chunk_sprites.contains(chunk_pos))
                {
                    chunk->touched = false;
                    auto* target = new JGL::RenderTarget({Core::Chunk::ChunkSize, Core::Chunk::ChunkSize}, {0, 0, 0, 0});
                    RenderChunkTexture(chunk_pos, target, chunk);
                    cached_chunk_sprites.insert({chunk_pos, target});
                }
                // rendertarget needs updating.
                else if (chunk->touched) {
                    // TODO: Modify RenderTarget in place.
                    chunk->touched = false;

                    auto* target = cached_chunk_sprites[chunk_pos];
                    RenderChunkTexture(chunk_pos, target, chunk);


                    //cached_chunk_sprites.erase(chunk_pos);
                    //auto* target = new JGL::RenderTarget({Core::Chunk::ChunkSize, Core::Chunk::ChunkSize}, {0,0,0,0});
                    //RenderChunkTexture(chunk_pos, target, chunk);
                    //cached_chunk_sprites.insert({chunk_pos, target});

                }

            }
        }
    }

    Color4 LocalWorld::GetSkyColorInterpolatedForTimeOfDay(float time) {
        float rounded_low = Math::Floor(time);
        float rounded_high = Math::Ceil(time);

        float delta = (time - rounded_low);

        Color4 color_low = GetSkyColorBaseForTimeOfDay(rounded_low);
        Color4 color_high = GetSkyColorBaseForTimeOfDay(rounded_high);

        return Color4::Lerp(color_low, color_high, delta);
    }
    Color4 LocalWorld::GetSkyColorBaseForTimeOfDay(float time) {
        if (time >= 23*60)
            return Colors::Black;
        else if (time >= 22*60)
            return Colors::Black;
        else if (time >= 21*60)
            return Colors::White;
        else if (time >= 20*60)
            return Colors::White;
        else if (time >= 19*60)
            return Colors::White;
        else if (time >= 18*60)
            return Colors::White;
        else if (time >= 17*60)
            return Colors::White;
        else if (time >= 16*60)
            return Colors::White;
        else if (time >= 15*60)
            return Colors::White;

        return Colors::White;
    }

    void LocalWorld::DrawSky() {

        Vector2 viewport_topleft = camera.ScaledViewport().minPoint;
        Vector2 viewport_bottomright = camera.ScaledViewport().maxPoint;

        Color4 base_bg_color = Colors::Blues::CornflowerBlue;

        float space_starts_at = -2000.f;

        float depth = camera.Position().y;

        // Draw space bg.
        if (depth < -space_starts_at) {
            float t = Math::Clamp(  -((camera.Position().y+space_starts_at) / space_starts_at), 0.f, 1.f);
            base_bg_color = base_bg_color.Lerp(Colors::Blues::MidnightBlue, t);
        }

        float deep_starts_at = 2000.f;

        // Draw deep bg.
        if (depth > deep_starts_at) {
            float t = Math::Clamp((camera.Position().y-deep_starts_at) / deep_starts_at, 0.f, 1.f);
            base_bg_color = base_bg_color.Lerp(Colors::Black, t);
        }


        glClearColor(base_bg_color.RN(), base_bg_color.GN(), base_bg_color.BN(),1.f);

        //if ()

    }

    void LocalWorld::Draw() {
        CheckCachedChunkSprites();

        JGL::J2D::Begin();


        //glClearColor(Colors::Blues::SkyBlue.RN(), Colors::Blues::SkyBlue.GN(), Colors::Blues::SkyBlue.BN(),1.f);

        DrawSky();

        // Shift the origin to the center of the screen.
        glTranslatef(camera.HalfSizeOffset().x, camera.HalfSizeOffset().y, 0);

        //DrawSky();

        // Apply rotation, zoom, and translation.
        glRotatef(camera.Rotation(), 0, 0, 1);
        glScalef(camera.Zoom(), camera.Zoom(), 1);

        glTranslatef(-camera.Position().x, -camera.Position().y, 0);


        // Draw the cached RenderTargets for our chunks.
        for (const auto& [chunk_pos, chunk] : loaded_chunks)
        {
            if (IsChunkCellWithinViewport(chunk_pos))
                RenderChunk(chunk_pos);

            //JGL::J2D::DrawString(Colors::Black, std::format("{}, {}", chunk_pos.x, chunk_pos.y), chunk.GetChunkRealCoordinates().x, chunk.GetChunkRealCoordinates().y,
            //                     1, 8, font);
        }

        // Debug Grid
        //DrawChunkGrid();



        // Banana for scale.
        JGL::J2D::FillRect(Colors::Blue, {0,0}, {64, 64});
        JGL::J2D::FillRect(Colors::Green, {0,0}, {32, 32});
        JGL::J2D::FillRect(Colors::Yellow, {0,0}, {16, 16});
        JGL::J2D::FillRect(Colors::Red, {0,0}, {8, 8});
        JGL::J2D::FillRect(Colors::White, {0,0}, {1, 1});

        for (auto* entity : entities) {
            entity->Draw();
        }


        JGL::J2D::End();


        //hud->Draw();
    }

    unsigned int LocalWorld::GetRenderTargetCount() const {
        return cached_chunk_sprites.size();
    }

    bool LocalWorld::IsChunkCellWithinViewport(const Vector2 &coords, int extraChunkRadius) const {
        int extraChunkComputedPixelsRequired = extraChunkRadius*Core::Chunk::ChunkSize;
        AABB2D chunk_bounding_box = AABB2D(coords*Core::Chunk::ChunkSize, coords*Core::Chunk::ChunkSize+Vector2(Core::Chunk::ChunkSize));
        return Core::Solver::AABB2Dvs(
                camera.ScaledViewportOversized(extraChunkComputedPixelsRequired), chunk_bounding_box);
    }

    RNG rng;



    void LocalWorld::RenderChunkTexture(const Vector2 &coords, JGL::RenderTarget* destination, Core::Chunk* chunk)
    {

#define DEBUG_TILE_UPDATES

        using CaveGame::Core::TileID;

        TileID t_id;
        Core::Tile* t_data;
        Color4 t_color;


        //std::vector<Vector2> stone_coords;
        //std::vector<Vector2> dirt_coords;
        //std::vector<Vector2> grass_coords;
        //std::vector<Vector2> clay_coords;
        //std::vector<Vector2> mud_coords;
        JGL::J2D::Begin(destination, true);
        for (int x = 0; x < Core::Chunk::ChunkSize; x++)
        {
            for (int y = 0; y < Core::Chunk::ChunkSize; y++)
            {
                t_id = chunk->GetTile(x, y);

                Vector2 relative_tile_coords = Vector2(x, y);
                const Vector2& tile_coords = relative_tile_coords;
                int wx = (x*Core::Chunk::ChunkSize) + x;
                int wy = (y*Core::Chunk::ChunkSize) + y;


#ifdef DEBUG_TILE_UPDATES
                if (chunk->GetTileUpdateFlag(x, y))
                    JGL::J2D::DrawPoint(Color4(255, 0, 0, 128), tile_coords);

                if (chunk->GetTileUpdateBufferFlag(x, y))
                    JGL::J2D::DrawPoint(Color4(255, 0, 0, 128), tile_coords);
#endif


                if (t_id == TileID::AIR || t_id == TileID::VOID) // Air
                    continue;




                t_data = Core::GetByNumeric(t_id);

                if (t_id == TileID::COBBLESTONE) {
                    float val = generator.Perlin(wx, wy, 12, 12, 12, 2);

                    if (val > 0.60f || val < -0.60f)
                        t_color = Core::Tiles::Cobblestone.color_pallet[1];
                    else if (val > 0.40f || val < -0.40f)
                        t_color = Core::Tiles::Cobblestone.color_pallet[0];
                    else if (val > 0.2f || val < -0.2f)
                        t_color = Core::Tiles::Cobblestone.color_pallet[2];
                    else
                        t_color = Core::Tiles::Cobblestone.color_pallet[3];
                    //Core::Tiles::Cobblestone.color_pallet
                } else if (t_data->has_color_pallet) {
                    uint rand = generator.ColorMap(t_data->color_pallet.size(), x, y);
                    t_color = t_data->color_pallet[rand];
                } else {
                    t_color = t_data->base_color;
                }

                JGL::J2D::DrawPoint(t_color, tile_coords);

            }
        }
        JGL::J2D::End();
    }

    Vector2 GetChunkRealCoordinates(const Vector2& cell) {
        return cell * Core::Chunk::ChunkSize;
    }

    void LocalWorld::RenderChunk(const Vector2& coords) {
        auto it = cached_chunk_sprites.find(coords);
        if (it != cached_chunk_sprites.end())
            JGL::J2D::DrawRenderTarget(it->second, GetChunkRealCoordinates(coords));
    }

    void LocalWorld::LookForChunksNeedUnloading() {
        for (auto it = loaded_chunks.begin(); it != loaded_chunks.end();)
        {
            const auto coords = it->first;
            if (!IsChunkCellWithinViewport(coords, 2)) {
                // TODO: Move off main thread.
                SaveChunkToFile(coords, it->second);

                delete it->second;
                loaded_chunks.erase(it++);

                auto sprite_it = cached_chunk_sprites.find(coords);
                if (sprite_it != cached_chunk_sprites.end())
                    delete sprite_it->second, cached_chunk_sprites.erase(sprite_it);
            }
            else
                ++it;
        }
    }

    //void LocalWorld::LookForChunksNeedLoading()
    void LocalWorld::LookForChunksNeedLoading()
    {
        Vector2 viewport_topleft = camera.ScaledViewportOversized(2*Core::Chunk::ChunkSize).minPoint;
        Vector2 viewport_bottomright = camera.ScaledViewportOversized(2*Core::Chunk::ChunkSize).maxPoint;

        int lower_bound_h = Math::Floor(viewport_topleft.x / Core::Chunk::ChunkSize)/*-2*/;
        int upper_bound_h = Math::Floor(viewport_bottomright.x / Core::Chunk::ChunkSize)/*+2*/;

        int lower_bound_v = Math::Floor(viewport_topleft.y / Core::Chunk::ChunkSize)/*-2*/;
        int upper_bound_v = Math::Floor(viewport_bottomright.y / Core::Chunk::ChunkSize)/*+2*/;

        for (int x = lower_bound_h; x <= upper_bound_h; x++)
        {
            for (int y = lower_bound_v; y <= upper_bound_v; y++)
            {
                Vector2 cell(x, y);
                if (!HasChunkAtCell(cell) && !AwaitingChunkAtCell(cell))
                    RequestChunk(cell);
            }
        }
    }

    void LocalWorld::Update(float elapsed) {
        World::Update(elapsed);

        camera.Update(elapsed);

        check_chunks_timer += elapsed;

        // TODO: Move this block to World class

        if (check_chunks_timer > 1/8.f) {
            // TODO: These procedures need to be performed asynchronously.
            check_chunks_timer = 0.f;
            LookForChunksNeedUnloading();
            LookForChunksNeedLoading();
            //auto a1 = std::async(&LocalWorld::LookForChunksNeedLoading, this);

            // TODO: Unload chunks once they're offscreen for a certain time threshold.


            // Check our generator queue for complete chunks, and pull them into our loaded chunks.
            while (!ServedChunks.empty())
            {
                Core::Chunk* c = nullptr;
                ServedChunks.front_pop(c);
                loaded_chunks.emplace(c->GetChunkCell(), c);

                if (std::find(chunks_in_waiting.begin(), chunks_in_waiting.end(), c->GetChunkCell()) != chunks_in_waiting.end())
                    chunks_in_waiting.erase(std::remove(chunks_in_waiting.begin(), chunks_in_waiting.end(), c->GetChunkCell()), chunks_in_waiting.end());

                //generator.SecondPass(this, c);
            }
        }
    }




    void LocalWorld::RefreshAll() {
        //World::RefreshAll();
        cached_chunk_sprites.clear();
        chunks_in_waiting.clear();
    }


}

