#pragma once

#include <Client/AssetService.hpp>

#include <J3ML/Geometry.hpp>
#include <Client/Splash.hpp>
#include <Client/MainMenu.hpp>
#include <Client/GameSession.hpp>
#include <Client/WindowWidgetManager.hpp>
#include <JUI/Widgets/ListLayout.hpp>
#include <JUI/UDim.hpp>
#include <rewindow/types/window.h>
#include <JUI/Widgets/Window.hpp>
#include <Client/Scene.hpp>
#include <Client/SceneManager.hpp>
#include <Client/Console.hpp>

namespace CaveGame::ClientApp
{

    using CaveGame::Client::Scene;
    using CaveGame::Client::SceneManager;

    /// The main program class. Everything originates here.
    /// This class is derived from RWindow class.
    class CaveGameWindow : public ReWindow::RWindow, public SceneManager
    {
    public:
        [[nodiscard]] bool InGameSession() const;
        [[nodiscard]] bool InMainMenu() const;
        Client::AssetService assets;
        CaveGame::Client::Splash* splash_ctx;
        CaveGame::Client::MainMenu* menu_ctx;
        CaveGame::Client::GameSession* game_ctx;
        JUI::Scene* wm;
        JUI::Window* settings_window;
        JUI::Window* credits_window;
        Client::Console* console_window;
        JUI::Window* stats_window;
        //Scene* current_scene = nullptr;
        bool render_grid = true;
        bool generate_grid = true;
        bool mbd = false;
        float our_avg = 0.f;

        /// Constructs and returns an instance of the game program.
        /// @param title
        CaveGameWindow(const std::string& title, int width, int height);

        ~CaveGameWindow();

        bool InGame() const;

        /// This function runs the totality of the game loop, start to finish.
        /// @note The game loop runs until Die() is called, at which point final cleanup is performed.
        void Run();

        /// This function triggers an internal signal that the game window is ready to close.
        /// @note This does not immediately close the window, rather, time is given to allow for final data saving and other processes.
        void Die();



        void create_console_window();

        void create_stats_window();

        void create_settings_window();

        void create_window_widgets();

        /// This function sets up the initial program state, particularly that which must be performed **after** the window is opened.
        void Init();
        void tile_draw(int x, int y, int radius, Core::TileID tile);
        /// This function cleans up any data or assets before closing the game.
        void Cleanup();

        /// This function performs logic calculation routines, once per refresh.
        void Update(float elapsed);

        /// This function performs rendering routines, once per refresh.
        void Draw();

    public:
        void OnRefresh(float elapsed) override;
        void OnMouseButtonDown(const ReWindow::MouseButtonDownEvent &ev) override;
        void OnMouseButtonUp(const ReWindow::MouseButtonUpEvent &ev) override;
        void OnKeyUp(const ReWindow::KeyUpEvent &ev) override;
        void OnKeyDown(const ReWindow::KeyDownEvent& ev) override;
        void OnMouseMove(const ReWindow::MouseMoveEvent &ev) override;
        bool OnResizeRequest(const ReWindow::WindowResizeRequestEvent &ev) override;

        void OnClosing() override;
    protected:
        /// Draws a sequence of 'debug' information strings to the screen, in a descending list.
        void draw_debug_info(std::vector<std::string> tokens);
        void CreateMenuWindows();
        /// Constructs the Splash screen, Main Menu screen, and In-game session.
        void CreateContexts();
        void Gameloop();
        void Step();

        bool wanna_die = false; // This field indicates the program is ready to close.
        void InGameControls(float elapsed);

        float tool_radius = 8.f;

        void OnMouseWheel(const ReWindow::MouseWheelEvent &) override;
    };
}