#include <Client/CreditsWindow.hpp>
#include "ClientApp/CaveGameWindow.hpp"

#include <bits/random.h>
#include <Core/Explosion.hpp>
#include <Core/Loggers.hpp>
#include <Core/Player.hpp>
#include <rewindow/inputservice.hpp>
#include <Core/Macros.hpp>

namespace CaveGame::ClientApp
{
    CaveGameWindow::CaveGameWindow(const std::string& title, int width, int height): ReWindow::RWindow(title, width, height)
    {
        Logs::Info("Creating game window.");

        CreateContexts();
        CreateMenuWindows();
    }

    CaveGameWindow::~CaveGameWindow() {

        Logs::Info("Removing game objects.");
        delete splash_ctx;
        delete menu_ctx;
        delete game_ctx;
        delete wm;

        Logs::Info("Closing game window.");
        // The base destructor will get called automatically - Redacted.
        //RWindow::~RWindow();
    }

    void CaveGameWindow::CreateContexts()
    {
        Logs::Info("Building splash screen.");
        splash_ctx = new CaveGame::Client::Splash();
        Logs::Info("Building main menu.");
        menu_ctx = new CaveGame::Client::MainMenu();
        game_ctx = new CaveGame::Client::GameSession();

        menu_ctx->RequestWorld += [this](Client::SingleplayerSessionInfo info) {
            Logs::Info(std::format("Game session requested.", info.NewWorld));
            game_ctx = new CaveGame::Client::GameSession(info.NewWorld);
            // TODO: Parse Info to construct gameworld files.
            ChangeScene(game_ctx);
        };
        menu_ctx->RequestQuit += [this]() {
            Die();
        };
        menu_ctx->RequestShowCredits += [this]()
        {
            credits_window->Visible(true);
        };
    }

    void CaveGameWindow::CreateMenuWindows()
    {
        wm = new JUI::Scene();
    }

    void CaveGameWindow::Step()
    {
        this->ManagedRefresh();
    }

    void CaveGameWindow::Gameloop()
    {
        while (!this->wanna_die) {
            Step();
        }
    }

    void CaveGameWindow::Run()
    {
        this->SetRenderer(RenderingAPI::OPENGL);
        this->Open();
        this->Init();
        this->SetResizable(true);
        this->SetVsyncEnabled(false);

        // TODO: Implement Resource/Asset manager rather than passing asset references around like this.
        this->menu_ctx->BuildWidgets();

        this->assets.TempLoadSimple();
        this->assets.EnqueueDefaultAssetsFolder();
        this->ChangeScene(this->splash_ctx);

        Gameloop();

        processOnClose();
        Cleanup();

        //DestroyWindow();
    }

    void CaveGameWindow::create_stats_window()
    {

    }

    void CaveGameWindow::create_settings_window()
    {
        settings_window = new JUI::Window(wm);
        settings_window->SetTitleFont(JGL::Fonts::Jupiteroid);
        settings_window->SetTitle("Settings");
        settings_window->Visible(false);
    }

    void CaveGameWindow::create_window_widgets()
    {
        create_console_window();
        credits_window = Client::CreateCreditsWindowWidget(wm);
        create_stats_window();
        create_settings_window();
    }

    void CaveGameWindow::Init()
    {
        // Load OpenGL
        gladLoadGL();

        // Init Josh Graphics
        bool result = JGL::Init(GetSize(), 0.f, 0.f);
        //JGL::InitTextEngine();
        JGL::Update(GetSize());

        // Fetch core assets
        // TODO: Asset manager? Need to share references and pointers of assets between many files..

        // Setup base OpenGL state
        glClearColor(0.f, 0.f, 0.f, 0.f);
        glEnable(GL_DEPTH_TEST);
        glDepthFunc(GL_LESS);
        glDepthMask(GL_TRUE);

        // Construct menu elements
        create_window_widgets();
    }

    void CaveGameWindow::tile_draw(int x, int y, int radius, Core::TileID tile)
    {

        game_ctx->world->SetTile(x, y, tile);

        for (int dx = -radius; dx <= radius; ++dx)
        {
            for (int dy = -radius; dy <= radius; ++dy)
            {
                if (Math::Abs(dx)+Math::Abs(dy) < radius)
                    game_ctx->world->SetTile(x+dx, y+dy, tile);
            }
        }
    }

    void CaveGameWindow::draw_debug_info(std::vector<std::string> tokens)
    {
        int font_size = 14;
        Color4 bg_color = {0,0,0,192};
        Color4 text_color = Colors::Blues::SkyBlue;
        Vector2 pos = {2, 2};
        Vector2 text_size = JGL::Fonts::Jupiteroid.MeasureString(tokens[0], font_size);

        JGL::J2D::Begin();
        int index = 0;

        for (auto token : tokens)
        {
            text_size = JGL::Fonts::Jupiteroid.MeasureString(token, font_size);
            JGL::J2D::FillRect(bg_color, pos, text_size);
            JGL::J2D::DrawString(text_color, token, pos.x, pos.y, 1.f, font_size);
            pos.y += text_size.y;
        }
        JGL::J2D::End();
    }

    void CaveGameWindow::InGameControls(float elapsed)
    {
        // TODO: Ideally, we want LocalWorld class to manage it's camera input.
        // however, we cannot use the "polling" set of input methods in the Context, GameSession, or LocalWorld classes,
        // so we had to bring this code up here.
        // @maxi suggests function pointers and closure.

        /*
        if (IsKeyDown(Keys::LeftArrow))
            game_ctx->world->camera.MoveLeft();
        if (IsKeyDown(Keys::RightArrow))
            game_ctx->world->camera.MoveRight();
        if (IsKeyDown(Keys::UpArrow))
            game_ctx->world->camera.MoveUp();
        if (IsKeyDown(Keys::DownArrow))
            game_ctx->world->camera.MoveDown();

        if (IsKeyDown(Keys::LeftBracket))
            game_ctx->world->camera.Rotate(-5*elapsed);
        if (IsKeyDown(Keys::RightBracket))
            game_ctx->world->camera.Rotate(5*elapsed);

        if (IsKeyDown(Keys::Minus))
            game_ctx->world->camera.ZoomIn(-elapsed);
        if (IsKeyDown(Keys::Equals))
            game_ctx->world->camera.ZoomIn(elapsed);
        */

        if (IsMouseButtonDown(MouseButtons::Left))
        {

            Vector2 transformed = game_ctx->world->camera.ScreenToWorld(GetMouseCoordinates()) - Vector2{0.5f, 0.5f};
            tile_draw(transformed.x, transformed.y, tool_radius, Core::TileID::AIR);
        }

        if (IsMouseButtonDown(MouseButtons::Right))
        {

            Vector2 transformed = game_ctx->world->camera.ScreenToWorld(GetMouseCoordinates()) - Vector2{0.5f, 0.5f};

            tile_draw(transformed.x, transformed.y, tool_radius, game_ctx->GetHotbarTile());

            /*game_ctx->world->SetTile(transformed.x, transformed.y, Core::TileID::GRASS);
            game_ctx->world->SetTile(transformed.x+1, transformed.y, Core::TileID::GRASS);
            game_ctx->world->SetTile(transformed.x-1, transformed.y, Core::TileID::GRASS);
            game_ctx->world->SetTile(transformed.x, transformed.y+1, Core::TileID::GRASS);
            game_ctx->world->SetTile(transformed.x, transformed.y-1, Core::TileID::GRASS);*/
        }
    }

    void CaveGameWindow::Update(float elapsed)
    {

        // TODO: We can't tell the game to change scenes from within a scene, currently.
        if (current_scene == splash_ctx && splash_ctx->SplashComplete())
            ChangeScene(menu_ctx);

        // Update Current Scene
        if (current_scene != nullptr)
        {
            current_scene->PassWindowSize(GetSize());
            current_scene->Update(elapsed);
            // TODO: current_scene->PassInputState(...); ?
        }

        // Update floating windows.
        wm->Update(elapsed);


        if (current_scene == game_ctx)
        {
            game_ctx->world->mouse_pos = GetMouseCoordinates();
            InGameControls(elapsed);
        }

    }

    void CaveGameWindow::Draw()
    {
        JGL::Update(GetSize());

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        // Draw Current Scene
        if (current_scene != nullptr)
            current_scene->Draw();


        // Draw floating windows.
        wm->Draw();

        std::vector<std::string> debug_lines =
            {
            "Re-CaveGame - Redacted Software",
            std::format("frame: {}", refresh_count),
            std::format("delta: {}ms", J3ML::Math::Round(delta_time*1000)),
            std::format("fps: {}", J3ML::Math::Round(refresh_rate)),
            std::format("avg: {}", J3ML::Math::Round(avg_refresh_rate))
        };


        if (current_scene == game_ctx) {

            Core::TileID id = game_ctx->GetTileUnderMouse();

            Core::Tile* data = Core::GetByNumeric(id);

            debug_lines.push_back(std::format("tile: {} id: {}", data->MnemonicID(), (uint)data->NumericID()));
            debug_lines.push_back(std::format("entities: {}", game_ctx->world->GetCurrentEntityCount()));
            debug_lines.push_back(std::format("tile_simulation: {}", game_ctx->world->GetTileSimulationEnabled()));
            debug_lines.push_back(std::format("rendertargets: {}", game_ctx->world->GetRenderTargetCount()));
        }

        draw_debug_info(debug_lines);

        if (current_scene == game_ctx) {

            JGL::J2D::Begin();

            auto camera = game_ctx->world->camera;

            // Shift the origin to the center of the screen.
            glTranslatef(camera.HalfSizeOffset().x, camera.HalfSizeOffset().y, 0);

            //DrawSky();

            // Apply rotation, zoom, and translation.
            glRotatef(camera.Rotation(), 0, 0, 1);
            glScalef(camera.Zoom(), camera.Zoom(), 1);

            glTranslatef(-camera.Position().x, -camera.Position().y, 0);

            auto pos = camera.ScreenToWorld(currentMouse.Position);

            JGL::J2D::OutlineCircle(Colors::Red, pos, tool_radius);

            JGL::J2D::DrawPoint({128, 128, 128, 128}, pos);

            JGL::J2D::End();
        }
    }



    void CaveGameWindow::OnRefresh(float elapsed)
    {

        Update(elapsed);
        Draw();
        int glError = glGetError();
        if (glError != GL_NO_ERROR) {
            Logs::Error(std::format("OpenGL Error: {}", glError));
        }
        GLSwapBuffers();

    }

    void CaveGameWindow::OnKeyUp(const ReWindow::KeyUpEvent& ev) {
        if (current_scene != nullptr)
            current_scene->PassKeyInput(ev.key, false);

        if (ev.key == Keys::G)
            generate_grid = !generate_grid;

        if (ev.key == Keys::V)
            render_grid = !render_grid;

        if (ev.key == Keys::Grave) {

            // TODO: Key Input Hierarchy: Determine and handle priority of where certain keys go.
            console_window->Visible(!console_window->IsVisible());
            return;
        }

        //RWindow::OnKeyUp(ev);

        wm->ObserveKeyInput(ev.key, false);
    }

    void CaveGameWindow::OnKeyDown(const ReWindow::KeyDownEvent &ev) {

        if (current_scene == game_ctx) {

            if (ev.key == Keys::F5)
                game_ctx->world->RefreshAll();

            if (ev.key == Keys::F7)
                game_ctx->world->SetTileSimulationEnabled(!game_ctx->world->GetTileSimulationEnabled());

            if (ev.key == Keys::F6) {
                std::srand(std::time(nullptr));
                game_ctx->world->SetSeed(std::rand());
            }

            if (ev.key == Keys::Escape) {
                game_ctx->SaveAndExit();
                ChangeScene(menu_ctx);
            }

            if (ev.key == Keys::P) {
                auto coords = game_ctx->world->camera.ScreenToWorld(InputService::GetMousePosition());
                auto* plr = new Core::Explosion(game_ctx->world, coords, 1.f, tool_radius);
                //auto* plr2 = new Core::Player(coords);
                //game_ctx->world->AddEntity(plr2);
                game_ctx->world->AddEntity(plr);
            }
        }

        if (ev.key == Keys::Grave)
            return;

        if (current_scene != nullptr)
            current_scene->PassKeyInput(ev.key, true);

        wm->ObserveKeyInput(ev.key, true);

    }

    void CaveGameWindow::OnMouseButtonDown(const ReWindow::MouseButtonDownEvent& ev)
    {

        if (current_scene != nullptr)
            current_scene->PassMouseInput(ev.Button.ButtonIndex, true);

        if (ev.Button == MouseButtons::Left)
            wm->ObserveMouseInput(JUI::MouseButton::Left, true);
        if (ev.Button == MouseButtons::Right)
            wm->ObserveMouseInput(JUI::MouseButton::Right, true);
        if (ev.Button == MouseButtons::Middle)
            wm->ObserveMouseInput(JUI::MouseButton::Middle, true);


        mbd = true;


        RWindow::OnMouseButtonDown(ev);
    }


    void CaveGameWindow::OnMouseButtonUp(const ReWindow::MouseButtonUpEvent& ev)
    {
        if (current_scene != nullptr)
            current_scene->PassMouseInput(ev.Button.ButtonIndex, false);

        if (ev.Button == MouseButtons::Left)
            wm->ObserveMouseInput(JUI::MouseButton::Left, false);
        if (ev.Button == MouseButtons::Right)
            wm->ObserveMouseInput(JUI::MouseButton::Right, false);
        if (ev.Button == MouseButtons::Middle)
            wm->ObserveMouseInput(JUI::MouseButton::Middle, false);


        mbd = false;

        RWindow::OnMouseButtonUp(ev);
    }

    void CaveGameWindow::OnMouseMove(const ReWindow::MouseMoveEvent& ev)
    {

        if (current_scene != nullptr)
            current_scene->PassMouseMovement(ev.Position);

        wm->ObserveMouseMovement(ev.Position);

        RWindow::OnMouseMove(ev);
    }

    bool CaveGameWindow::OnResizeRequest(const ReWindow::WindowResizeRequestEvent& ev)
    {

        return true;
    }

    void CaveGameWindow::OnMouseWheel(const ReWindow::MouseWheelEvent &ev) {
        tool_radius -= ev.WheelMovement/2.f;


        tool_radius = Math::Max(tool_radius, 0.45f);
    }

    void CaveGameWindow::Die() { wanna_die = true; }

    bool CaveGameWindow::InGameSession() const {
        return (current_scene == game_ctx);
    }

    bool CaveGameWindow::InMainMenu() const {
        return (current_scene == menu_ctx);
    }

    void CaveGameWindow::Cleanup() {

    }

    void CaveGameWindow::OnClosing() {
        if (current_scene == game_ctx)
        {
            game_ctx->SaveAndExit();
        }

        wanna_die = true;
    }

    void CaveGameWindow::create_console_window() {
        console_window = new Client::Console(this->wm);
        console_window->Visible(false);

        console_window->OnInput += [this] (const std::string& msg)
        {
            auto tokens = string_split(msg, ' ');
            if (tokens[0] == "tp")
            {
                std::cout << "Teleport Command" << std::endl;
            }

        };
    }

    bool CaveGameWindow::InGame() const {
        return current_scene == game_ctx;
    }

}


