/// CaveGame - A procedural 2D platformer sandbox.
/// Created by Josh O'Leary @ Redacted Software, 2020-2024
/// Contact: josh@redacted.cc
/// Contributors: william@redacted.cc maxi@redacted.cc
/// This work is dedicated to the public domain.

/// @file Chunk.hpp
/// @desc A subdivision of the game-world into discrete blocks.
/// @edit 12/2/2024

#pragma once

#include <J3ML/Geometry/QuadTree.hpp>
#include "Tile.hpp"
#include "Data.hpp"
#include "Serialization.hpp"
#include <filesystem>
#include <fstream>

namespace CaveGame::Core
{


    class Chunk {
    public:
        static constexpr int ChunkSize = 128;
    public:
        /// The default constructor does not initialize any members.
        Chunk() = default;
        ~Chunk() = default;


        /// Constructs a chunk with empty (air) tiles, located at the given chunk-cell.
        /// @param cell_coords Specifies the chunk-cell this chunk will occupy.
        explicit Chunk(const Vector2 &cell);

        /// Constructs a chunk from a filesystem path. This constructor will load and parse the chunk file.
        explicit Chunk(const Vector2 &cell, const std::filesystem::path &file);

        /// Returns the TileID enumeration that occupies the given tile-cell.
        TileID GetTile(int x, int y) const;

        /// Sets the tile ID at the given tile-cell.
        /// @param trigger_tile_updates Whether to force adjacent tiles to update.
        void SetTile(int x, int y, TileID t, bool trigger_tile_updates = true);

        /// Sets the tile ID at the given tile-cell, and explicitly does not update adjacent tiles.
        void SetTileSilent(int x, int y, TileID t);

        /// Returns the value of the update flag field at the given tile-cell.
        bool GetTileUpdateFlag(int x, int y) const;

        bool GetTileUpdateBufferFlag(int x, int y) const;

        void SetTileUpdateBufferFlag(int x, int y, bool flag = true);

        /// Sets the value of the update flag field at the given tile-cell.A
        void SetTileUpdateFlag(int x, int y, bool flag = true);

        /// Returns the "chunk coordinates" that this chunk occupies.
        [[nodiscard]] Vector2 GetChunkCell() const;

        /// Returns the world-coordinates of the top-leftmost tile of this chunk.
        [[nodiscard]] Vector2 GetChunkRealCoordinates() const;

        [[nodiscard]] const TileID* ptr() const { return &tiles[0][0];}


        [[nodiscard]] std::array<TileID, ChunkSize * ChunkSize> DataContiguous() const;

        void write(Buffer& buf) {
            write_f32(buf, cell.x);
            write_f32(buf, cell.y);

            for (int x = 0; x < ChunkSize; x++) {
                for (int y = 0; y < ChunkSize; y++) {
                    write_u8(buf, (uint8_t)GetTile(x, y));
                }
            }
        }

        void read(Buffer& buf)
        {
            cell.x = read_f32(buf);
            cell.y = read_f32(buf);

            for (int x = 0; x < ChunkSize; x++)
            {
                for (int y = 0; y < ChunkSize; y++)
                {
                    SetTile(x, y, (TileID) read_u8(buf));
                }
            }
        }

        void SetData(char* buffer);

        [[nodiscard]] static constexpr std::size_t BufferSizeBytes();

        [[nodiscard]] static constexpr std::size_t BufferSize();


        void SwapTileUpdateBuffers();

        bool FirstPassCompleted() const;

        bool SecondPassCompleted() const;

        void SetFirstPassCompleted(bool complete = true);

        void SetSecondPassCompleted(bool complete = true);

    public:
        bool touched;
    protected:
        Vector2 cell;
        TileID tiles[ChunkSize][ChunkSize];
        bool tagged_for_update[ChunkSize][ChunkSize];
        bool update_buffer[ChunkSize][ChunkSize];
        bool first_pass_complete = false;
        bool second_pass_complete = false;

    };

    constexpr std::size_t Chunk::BufferSizeBytes()
    {
        return (ChunkSize*ChunkSize)*sizeof(TileID);
    }

    constexpr std::size_t Chunk::BufferSize()
    {
        return (ChunkSize*ChunkSize);
    }
}
