#pragma once

#include <Core/Data.hpp>
#include <Core/Chunk.hpp>
#include <Core/Simplex.hpp>
#include <Core/Perlin.hpp>
#include <Core/Octave.hpp>

namespace CaveGame::Core
{

    // TODO: Terrain Generator Window Widget

    class Generator
    {
    public:
        /// Defines the sample count for the pre-generated RNG lookup table.
        static constexpr int   PrecomputedWhiteNoiseResolution = 2048;
        /// Shifts the ground-level of terrain vertically by this many tiles.

#pragma region Surface Parameters
        static constexpr float BaseSurfaceLvl = 100.f;

        static constexpr float HeightMapHighPassAmplitude = 700.f;
        static constexpr float HeightMapHighPassScale = 700.f;
        static constexpr float HeightMapLowPassAmplitude = 85.f;
        static constexpr float HeightMapLowPassScale = 64.f;
        static constexpr float TopSoilDepth = 345.f;

#pragma endregion
#pragma region Cave Parameters

        static constexpr float CaveInputScale = 800.f;
        static constexpr float CaveOutputScale = 1.20f;
        static constexpr float CaveErosionRange = 0.080f;
        static constexpr float CaveWhiteNoise = 0.005f;
        static constexpr float CaveAdditiveInputScale = 80.f;
        static constexpr float CaveAdditiveOutputScale = 0.45f;
        static constexpr float SurfaceCaveShrinkDepth = 100;
        static constexpr float SurfaceCaveShrinkFactor = 1.5f;

#pragma endregion
#pragma region Ore Vein Parameters

        const Vector2 ClayVeinHiPassScale = {200.f, 205.f};
        static constexpr float   ClayVeinHiPassOffset = 0.f;
        static constexpr float   ClayVeinHiPassOutputScale = 2.2f;

        const Vector2 ClayVeinLoPassScale = {17.f, 15.f};
        static constexpr float   ClayVeinLoPassOffset = 420.f;
        static constexpr float   ClayVeinLoPassOutputScale = 1.2f;

        static constexpr float   ClayVeinNoise = 0.175f;
        static constexpr float   ClayVeinRampFactor = 2.f;

        const Vector2 SiltVeinHiPassScale = {220.f, 205.f};
        static constexpr float   SiltVeinHiPassOffset = 5.f;
        static constexpr float   SiltVeinHiPassOutputScale = 2.2f;

        const Vector2 SiltVeinLoPassScale = {27.f, 25.f};
        static constexpr float   SiltVeinLoPassOffset = 422.f;
        static constexpr float   SiltVeinLoPassOutputScale = 1.25f;

        static constexpr float   SiltVeinNoise = 0.175f;
        static constexpr float   SiltVeinRampFactor = 2.f;

        const Vector2 StoneVeinHiPassScale = {30.f, 30.f};
        static constexpr float StoneVeinHiPassOffset = 666.f;
        static constexpr float StoneVeinHiPassOutputScale = 2.4f;

        const Vector2 StoneVeinLoPassScale = {220.f, 220.f};
        static constexpr float StoneVeinLoPassOffset = 0.5f;
        static constexpr float StoneVeinLoPassOutputScale = 1.75f;

        static constexpr float StoneVeinNoise = 0.25f;
        static constexpr float StoneVeinRampFactor = 1.75f;

        const Vector2 DirtVeinHiPassScale = {30.f, 30.f};
        static constexpr float DirtVeinHiPassOffset = 12.f;
        static constexpr float DirtVeinHiPassOutputScale = 1.35f;

        const Vector2 DirtVeinLoPassScale = {90.f, 90.f};
        static constexpr float DirtVeinLoPassOffset = 0.5f;
        static constexpr float DirtVeinLoPassOutputScale = 0.75f;

        const float DirtVeinNoise = 0.05f;
        static constexpr float DirtVeinRampFactor = 0.5f;

        const Vector2 CoalVeinHiPassScale = {20.f, 20.f};
        static constexpr float CoalVeinHiPassOffset = 12.f;
        static constexpr float CoalVeinHiPassOutputScale = 1.05f;

        const Vector2 CoalVeinLoPassScale = {120.f, 120.f};
        static constexpr float CoalVeinLoPassOffset = 0.5f;
        static constexpr float CoalVeinLoPassOutputScale = 0.83f;

        static constexpr float CoalVeinNoise = 0.45f;
        static constexpr float CoalVeinRampFactor = 0.75f;

        const Vector2 CopperVeinHiPassScale = {30.f, 30.f};
        static constexpr float CopperVeinHiPassOffset = 12.f;
        static constexpr float CopperVeinHiPassOutputScale = 1.35f;

        const Vector2 CopperVeinLoPassScale = {90.f, 90.f};
        static constexpr float CopperVeinLoPassOffset = 0.5f;
        static constexpr float CopperVeinLoPassOutputScale = 0.75f;

        static constexpr float CopperVeinNoise = 0.05f;
        static constexpr float CopperVeinRampFactor = 0.5f;

        const Vector2 TinVeinHiPassScale = {30.f, 30.f};
        static constexpr float TinVeinHiPassOffset = 12.f;
        static constexpr float TinVeinHiPassOutputScale = 1.35f;

        const Vector2 TinVeinLoPassScale = {90.f, 90.f};
        static constexpr float TinVeinLoPassOffset = 0.5f;
        static constexpr float TinVeinLoPassOutputScale = 0.75f;

        static constexpr float TinVeinNoise = 0.05f;
        static constexpr float TinVeinRampFactor = 0.5f;

        const Vector2 IronVeinHiPassScale = {30.f, 30.f};
        static constexpr float IronVeinHiPassOffset = 12.f;
        static constexpr float IronVeinHiPassOutputScale = 1.35f;

        const Vector2 IronVeinLoPassScale = {90.f, 90.f};
        static constexpr float IronVeinLoPassOffset = 0.5f;
        static constexpr float IronVeinLoPassOutputScale = 0.75f;

        static constexpr float IronVeinNoise = 0.05f;
        static constexpr float IronVeinRampFactor = 0.5f;
#pragma endregion

    public:
        //Generator() = default;
        explicit Generator(int seed);

        int GetSeed() const { return seed;}
        void SetSeed(int newSeed) { seed = newSeed; }

        TileID HeightMap(float depth, int wx, int wy);
        float Perlin(int wx, int wy, float hScale, float vScale, float offset, float outputScale);

        float Octave(int wx, int wy, float hScale, float vScale, float offset, float outputScale, int octaves);

        /// Returns the sum of hi+lo and (hi*lo) / div, resulting in something approximating noise iterated in 2 octaves.
        float AddSumAndPartialProduct(float hi, float lo, float div);

        TileID HiLoSelect(float pass, float upperBound, float lowerBound, TileID hiSelect, TileID loSelect, TileID fallback);

        /// This function returns a floating point number (Usually in the range [-1, 1]) to be used as a condition for
        /// whether ores should spawn at this location, given a set of noise parameters.
        /// Ores are generated by computing two Perlin Noise passes, taking the sum of adding and multiplying, then adding in random noise.
        /// Hi-Pass is used to define macro (larger-scale) features of ore veins, and Lo-Pass adds minor detailing for a more organic appearance.
        float ComputeOre(int wx, int wy, const Vector2& hiPass, const Vector2& loPass, float hiOffset, float loOffset,
                         float hiScale, float loScale, float noise, float ramp);

        TileID ComputeTile(int wx, int wy);
        void FirstPass(Chunk* chunk);
        void SecondPass(ITileMap* world, Chunk* chunk);


        uint ColorMap(int range, int wx, int wy);


        // TODO: Implement SecondPass, the catch is, it **may** need to load an arbitrary amount of adjacent chunks to correctly place structures.
        // TODO: Expert Mode: How do we keep it threaded separately while still accomplishing the above?

        int ModInt(int x, int mod) const {

            x = Math::Abs(x);

            if (x == 0)
                return 0;

            return x % mod;
        }

        /// Returns a value between 0 and 1.
        float GetPrecomputedWhiteNoise1D(int x) const;

        /// Returns a value between 0 and 1.
        float GetPrecomputedWhiteNoise2D(int x, int y) const;

    protected:
        int seed = 0;
        SimplexNoise simplex;
        PerlinNoise perlin;

        RNG rng;
        float precomputed_white_noise_1D[PrecomputedWhiteNoiseResolution];
        float precomputed_white_noise_2D[PrecomputedWhiteNoiseResolution][PrecomputedWhiteNoiseResolution];
    private:
        float ComputeDepth(int wx, int wy);
    };
}