#include <map>
#include "Chunk.hpp"
#include "Generator.hpp"
#include "J3ML/LinearAlgebra/Vector2i.hpp"
#include "Interfaces.hpp"
#include <unordered_map>
#include <filesystem>
#include <future>
#include "ConcurrentQueue.hpp"
#include "Entity.hpp"


#pragma once


/// Implements a naiive-AF hashing algorithm for Vector2 class
/// So we can use it as unique keys in std::unordered_map
template<>
struct std::hash<Vector2>
{
    std::size_t operator()(const Vector2& k) const
    {
        return Math::Pow(k.x, 3) + Math::Pow(k.y, 2);
    }
};

namespace CaveGame::Core
{

    class WorldAsyncIOThreadManager
    {

    };

    /// The world class manages and coordinates terrain generation, chunk file IO, thread synchronization, and entity logic.
    class World : public ITileMap {
    public:
        constexpr static float TileTiccRate = 48.f; // Ticcs per second.
        constexpr static uint RandomTileTickCoefficient = 100;
        constexpr static float PeriodicAutosaveIntervalSeconds = 30.f;
        World() = default;
        ~World() = default;

        explicit World(const std::string &worldName, int seed = 0, bool overwrite = false);

        Vector2 GetChunkCoordinatesAtCell(int x, int y) const {
            float chunkX = Math::Floor((float)x / Chunk::ChunkSize);
            float chunkY = Math::Floor((float)y / Chunk::ChunkSize);

            return Vector2(chunkX, chunkY);
        }

        Vector2 GetTileCoordinatesAtCell(int x, int y) const {
            float tileX = Math::Mod(x, Chunk::ChunkSize);
            float tileY = Math::Mod(y, Chunk::ChunkSize);

            if (tileX < 0)
                tileX = Chunk::ChunkSize + tileX;
            if (tileY < 0)
                tileY = Chunk::ChunkSize + tileY;

            return Vector2(tileX, tileY);
        }

        TileID GetTile(int x, int y) const override;

        void SetTile(int x, int y, TileID t, bool flag_update = true) override;
        void SwapTile(int source_x, int source_y, int dest_x, int dest_y, bool flag_update = true) override;

        bool GetTileSimulationEnabled() const;

        void SetTileSimulationEnabled(bool enabled);

        int GetSeed() const {
            return generator.GetSeed();
        }
        void SetSeed(int seed) {
            generator.SetSeed(seed);
        }


        TileState GetTileState(int x, int y) const override { return 0; /* TODO: Implement tile state field */ }

        void SetTileState(int x, int y, TileState state) override { /* TODO: Implement tile state field */ }

        void SetTileUpdateFlag(int x, int y, bool flag) override;

        bool GetTileUpdateFlag(int x, int y) const override;

        // TODO: Doesn't really belong here.
        Vector2 ToUnitDirection(float rotation);

        Chunk* GetChunkAtCell(const Vector2 &cell);

        std::unordered_map<Vector2, Chunk*> GetChunkList();

        void DoRandomTileTicks();

        void DoForcedTileTicks();

        void DoRandomTileTick(const Vector2& coords, Chunk* chunk);

        void DoForcedTileTick(const Vector2& coords, Chunk* chunk);

        virtual void Update(float elapsed);

        virtual void RefreshAll();

        virtual void AutoSaveAsync() { }

        virtual void AutoSave();

        /// Gracefully saves all game state and closes the play session.
        virtual void SaveAndExit();

        /// Returns whether or not the world currently has a chunk loaded at the specified chunk coordinates.
        /// @note Valid Chunk coordinates are integers only.
        bool HasChunkAtCell(const Vector2 &cell) const;

        bool HasChunkOnFile(const Vector2 &cell) const;

        bool AwaitingChunkAtCell(const Vector2 &cell) const;

        void ChunkServerThread();

        std::filesystem::path GetWorldRootPath() const;

        std::filesystem::path GetChunkFullPath(const Vector2 &cell) const;


        /// Adds the entity pointer to the world, via std::move, this may leave the original entity pointer empty.
        void AddEntity(Entity* e);


        /// Returns the number of extant entities in this world right now.
        unsigned int GetCurrentEntityCount() const;

        /// Marks that the chunk at the given cell coordinates is needed.
        /// This will be picked up by the generator thread, and either generated, or loaded from disk.
        void RequestChunk(const Vector2& cell);
        void DropChunk(const Vector2& cell);

    protected:
        void SaveChunkToFile(const Vector2 &cell, Chunk* chunk);
        bool ValidCoords(const Vector2 &cell) const;
    protected:
        std::vector<Entity*> entities;
        const std::filesystem::path worlds {"worlds"};
        std::unordered_map<Vector2, Chunk*> loaded_chunks;
        Generator generator;
        float time_of_day;
        std::string world_name;
        RNG rng;
        float autosave_timer = 0.f;
        bool simulate_tiles = true;
        std::thread chunk_thread;
        std::atomic<bool> run_chunk_thread = true;
        ConcurrentQueue<Vector2> RequestedChunks;
        std::vector<Vector2> chunks_in_waiting;
        ConcurrentQueue<Core::Chunk*> ServedChunks;

        float tile_ticc_counter;

    private:



    };
}