#include <Core/Chunk.hpp>
#include "Core/Perlin.hpp"
#include "Core/Simplex.hpp"
#include "Core/Data.hpp"
#include <iostream>

namespace CaveGame::Core
{
    void Chunk::SetTile(int x, int y, TileID t, bool trigger_tile_updates) {
        if (!(0 <= x && x < ChunkSize && 0 <= y && y < ChunkSize))
            assert("Out of bounds!");

        if (tiles[x][y] != t)
        {
            tiles[x][y] = t;
            touched = true;

            if (trigger_tile_updates)
            {
                SetTileUpdateFlag(x, y);
            }
        }
    }

    void Chunk::SetTileSilent(int x, int y, TileID t)
    { SetTile(x, y, t, false); }

    bool Chunk::GetTileUpdateFlag(int x, int y) const
    {
        return tagged_for_update[x][y];
    }

    bool Chunk::GetTileUpdateBufferFlag(int x, int y) const {
        return update_buffer[x][y];
    }

    void Chunk::SetTileUpdateBufferFlag(int x, int y, bool flag) {
        update_buffer[x][y] = flag;
    }

    void Chunk::SetTileUpdateFlag(int x, int y, bool flag)
    {
        tagged_for_update[x][y] = flag;
    }

    TileID Chunk::GetTile(int x, int y) const {
        if (!(0 <= x && x < ChunkSize || 0 <= y && y < ChunkSize))
            throw std::runtime_error("Out of bounds!");
        return tiles[x][y];
    }

    Chunk::Chunk(const Vector2& cell_coords) : cell(cell_coords), touched(false), update_buffer{}, tagged_for_update{}, tiles{} {}

    Chunk::Chunk(const Vector2& cell_coords, const std::filesystem::path& file): Chunk(cell_coords)
    {
        std::ifstream inp;

        inp.open(file, std::ios::binary | std::ios::in);
        inp.seekg(0, std::ios::end);
        int length = inp.tellg();
        inp.seekg(0, std::ios::beg);

        char buffer[length];
        inp.read(buffer, length);
        inp.close();

        SetData(buffer);
    }

    Vector2 Chunk::GetChunkRealCoordinates() const {
        return GetChunkCell()*Core::Chunk::ChunkSize;
    }

    std::array<TileID, Chunk::ChunkSize * Chunk::ChunkSize> Chunk::DataContiguous() const {
        std::array<TileID, ChunkSize * ChunkSize> result{};
        memcpy(result.data(), tiles, ChunkSize * ChunkSize * sizeof(TileID));
        return result;
    }

    void Chunk::SetData(char* buffer) {
        auto* data = reinterpret_cast<TileID*>(buffer);
        memcpy(tiles, data, ChunkSize * ChunkSize * sizeof(TileID));
    }

    void Chunk::SwapTileUpdateBuffers() {
        memcpy(update_buffer, tagged_for_update, sizeof(update_buffer));
        std::memset(tagged_for_update, 0, sizeof(tagged_for_update));
    }

    bool Chunk::FirstPassCompleted() const { return first_pass_complete;}

    bool Chunk::SecondPassCompleted() const { return second_pass_complete;}

    void Chunk::SetFirstPassCompleted(bool complete) { first_pass_complete = complete; }

    void Chunk::SetSecondPassCompleted(bool complete) { second_pass_complete = complete; }

    Vector2 Chunk::GetChunkCell() const { return cell;}

}
