#include <Core/Loggers.hpp>
#include <Core/Tile.hpp>

namespace CaveGame::Core
{
    std::array<Tile*, 65536> registered_tiles{};


    void Tile::DecayTo(ITileMap *world, TileState state, int x, int y, TileID TDecaysTo) {
        world->SetTile(x, y, TDecaysTo);
    }

    bool Tile::ShouldSpread(ITileMap *world, int x, int y, TileID spreads_to) const {
        if (world->GetTile(x, y) == spreads_to)
        {
            if (world->HasAdjacentOrDiagonalAirBlock(x, y))
                return true;
        }
        return false;
    }

    LiquidTile::LiquidTile(): Tile() {}

    LiquidTile::LiquidTile(TileID id, const std::string &name, const Color4 &color, const std::vector<Color4> &pallet): Tile(id,name, color,pallet) {
        does_random_ticc = true;
    }

    LiquidTile::LiquidTile(TileID numeric, const std::string &name, Color4 color): Tile(numeric, name, color) {
        does_random_ticc = true;
    }

    SoilTile::SoilTile(): Tile() {}

    SoilTile::SoilTile(TileID id, const std::string &name, const Color4 &color, const std::vector<Color4> &pallet,
                       TileID decays_target): Tile(id,name, color,pallet) {
        decays_to = decays_target;
        does_random_ticc = true;
    }

    SoilTile::SoilTile(TileID numeric, const std::string &name, Color4 color, TileID decays_target): Tile(numeric, name, color) {
        decays_to = decays_target;
        does_random_ticc = true;
    }

    bool Tile::ShouldSuffocate(ITileMap *world, int x, int y) const {
        if (!world->HasAdjacentOrDiagonalAirBlock(x, y))
            return true;
        return false;
    }

    bool Tile::DecayCheck(ITileMap *world, TileState state, int x, int y, TileID decays_to) {
        if (ShouldSuffocate(world, x, y))
        {
            DecayTo(world, state, x, y, decays_to);
            return true;
        }
        return false;
    }

    bool SoilTile::SpreadCheck(ITileMap *world, TileState state, int x, int y, TileID spreads_to) {
        if (ShouldSpread(world, x, y+1, spreads_to)) {
            world->SetTile(x, y+1, numeric_id);
            return true;
        }
        if (ShouldSpread(world, x, y-1, spreads_to)) {
            world->SetTile(x, y-1, numeric_id);
            return true;
        }
        if (ShouldSpread(world, x+1, y, spreads_to)) {
            world->SetTile(x+1, y, numeric_id);
            return true;
        }
        if (ShouldSpread(world, x-1, y, spreads_to)) {
            world->SetTile(x-1, y, numeric_id);
            return true;
        }

        // Check diagonal blocks.
        if (ShouldSpread(world, x+1, y-1, spreads_to))
        {
            world->SetTile(x+1, y-1, numeric_id);
            return true;
        }

        if (ShouldSpread(world, x-1, y-1, spreads_to))
        {
            world->SetTile(x-1, y-1, numeric_id);
            return true;
        }

        if (ShouldSpread(world, x+1, y+1, spreads_to))
        {
            world->SetTile(x+1, y+1, numeric_id);
            return true;
        }

        if (ShouldSpread(world, x-1, y+1, spreads_to))
        {
            world->SetTile(x-1, y+1, numeric_id);
            return true;
        }

        return false;
    }

    void SoilTile::RandomTicc(ITileMap *world, TileState state, int x, int y) {
        if (!world->HasAdjacentOrDiagonalAirBlock(x, y))
            DecayTo(world, state, x, y, decays_to);
    }

    GrassyTile::GrassyTile(): SoilTile() {}

    GrassyTile::GrassyTile(TileID id, const std::string &name, const Color4 &color, const std::vector<Color4> &pallet,
        TileID decays_target): SoilTile(id,name, color,pallet, decays_target) { }

    GrassyTile::GrassyTile(TileID numeric, const std::string &name, Color4 color, TileID decays_target): SoilTile(numeric, name, color, decays_target) { }

    GrassyTile::GrassyTile(TileID id, const std::string &name, const Color4 &color, const std::vector<Color4> &pallet): SoilTile(id,name, color,pallet, TileID::DIRT) { }

    GrassyTile::GrassyTile(TileID numeric, const std::string &name, Color4 color): SoilTile(numeric, name, color, TileID::DIRT) { }

    void GrassyTile::RandomTicc(ITileMap* world, TileState state, int x, int y)
    {
        if (DecayCheck(world, state, x, y, decays_to))
            return;

        if (SpreadCheck(world, state, x, y, decays_to))
            return;

        if (world->GetTile(x, y+1) == TileID::AIR)
        {
            world->SetTile(x, y+1, TileID::VINE);
            return;
        }

    }

    MossyTile::MossyTile(): SoilTile()
    {}

    MossyTile::MossyTile(TileID id, const std::string& name, const Color4& color, TileID decays_target): SoilTile(id, name, color, decays_target) { }

    MossyTile::MossyTile(TileID id, const std::string &name, const Color4 &color, const std::vector<Color4> &pallet): SoilTile(id,name, color,pallet, TileID::STONE) { }

    MossyTile::MossyTile(TileID numeric, const std::string &name, Color4 color): SoilTile(numeric, name, color, TileID::STONE) { }

    void MossyTile::RandomTicc(ITileMap* world, TileState state, int x, int y)
    {
        if (DecayCheck(world, state, x, y, decays_to))
            return;

        if (SpreadCheck(world, state, x, y, decays_to))
            return;

    }

    GravityTile::GravityTile(): Tile()
    {
        //does_forced_ticc = true;
    }

    GravityTile::GravityTile(TileID id, const std::string& name, const Color4& color): Tile(id, name, color)
    {
        //does_forced_ticc = true;
    }

    GravityTile::GravityTile(TileID id, const std::string& name, const Color4& color, const std::vector<Color4>& pallet):Tile(id, name, color, pallet)
    {
        //does_forced_ticc = true;
    }

    void GravityTile::ForcedTicc(ITileMap *world, TileState state, int x, int y) {
        TileID below = world->GetTile(x, y + 1);
        if (below == TileID::AIR || below == TileID::WATER)
            world->SwapTile(x, y, x, y + 1);

        // TODO Change this to check specifically for gravity tile.
        else if (below == TileID::SAND && world->GetTile(x, y + 2) == TileID::SAND) {
            bool rng_roll = rand() % 2 == 0;
            if (world->GetTile(rng_roll ? (x + 1) : (x - 1), y) == TileID::AIR)
                world->SwapTile(x, y, rng_roll ? (x + 1) : (x - 1), y);
        }
    }

    VineTile::VineTile(): Tile()
    {
        //does_forced_ticc = true;
    }

    VineTile::VineTile(TileID id, const std::string& name, const Color4& color): Tile(id, name, color)
    {
        does_random_ticc = true;
    }

    VineTile::VineTile(TileID id, const std::string& name, const Color4& color, const std::vector<Color4>& pallet):Tile(id, name, color, pallet)
    {
        does_random_ticc = true;
    }

    void VineTile::ForcedTicc(ITileMap* world, TileState state, int x, int y)
    {
        if (DecayCheck(world, state, x, y, TileID::AIR))
            return;
    }

    void VineTile::RandomTicc(ITileMap* world, TileState state, int x, int y)
    {
        if (DecayCheck(world, state, x,y, TileID::AIR))
            return;

        if (SpreadCheck(world, state, x, y, TileID::AIR))
            return;
    }

    bool VineTile::DecayCheck(ITileMap* world, TileState state, int x, int y, TileID decays_to)
    {
        TileID above = world->GetTile(x, y-1);
        if (above != TileID::VINE && above != TileID::GRASS)
        {
            DecayTo(world, state, x, y, TileID::AIR);
            return true;
        }
        return false;
    }

    bool VineTile::ShouldSpread(ITileMap* world, int x, int y, TileID spreads_to) const
    {
        if (world->GetTile(x, y) == spreads_to)
        {
            return true;
        }
        return false;
    }

    bool VineTile::SpreadCheck(ITileMap* world, TileState state, int x, int y, TileID spreads_to)
    {
        if (ShouldSpread(world, x, y + 1, spreads_to))
        {
            if (world->GetTile(x+1, y+1) == TileID::AIR && world->GetTile(x-1, y+1) == TileID::AIR)
            {
                world->SetTile(x, y+1, numeric_id);
                return true;
            }
        }
        return false;
    }




    Tile *GetByName(const std::string &name) {
        // TODO: Optimize with additional mapping!!
        for(auto& tile : registered_tiles)
            if (tile->MnemonicID() == name)
                return tile;

        throw std::runtime_error("Invalid mnemonic ID!");
    }

    Tile::Tile() {}

    Tile::Tile(TileID id, const std::string &name, const Color4 &color): numeric_id(id), mnemonic_id(name), base_color(color),
                                                                         has_color_pallet(false) {
        RegisterTile(this);
    }

    Tile::Tile(TileID id, const std::string &name, const Color4 &color, const std::vector<Color4> &pallet): numeric_id(id), mnemonic_id(name), base_color(color), color_pallet(pallet),
        has_color_pallet(true) {
        RegisterTile(this);
    }

    TileID Tile::NumericID() const {  return numeric_id; }

    std::string Tile::MnemonicID() const {  return mnemonic_id; }

    Tile *GetByNumeric(TileID id) {
        // TODO: Optimize with additional mapping!!
        //if (registered_tiles.at((uint16_t)id))
            return registered_tiles[(uint16_t)id];
        //for(auto& tile : registered_tiles)
            //if (tile->numeric_id == id)
               //return tile;
        //throw std::runtime_error("ID not in tile registry!");
        return nullptr;
    }

    void RegisterTile(Tile *data) {
        uint16_t index = (uint16_t)data->NumericID();
        registered_tiles[index] = data;
    }

    WaterTile::WaterTile() : LiquidTile() {}

    WaterTile::WaterTile(TileID id, const std::string &name, const Color4 &color) : LiquidTile(id, name, color) {}
}